# sage_setup: distribution = sagemath-polyhedra
from sage.misc.lazy_import import lazy_import

lazy_import('sage.game_theory', 'catalog', 'game_theory')
lazy_import('sage.game_theory.cooperative_game', 'CooperativeGame')
lazy_import('sage.game_theory.normal_form_game', 'NormalFormGame')
lazy_import('sage.game_theory.matching_game', 'MatchingGame')
del lazy_import



# sage_setup: distribution = sagemath-polyhedra
"""
Co-operative Games With Finite Players

This module implements a class for a characteristic function cooperative
game. Methods to calculate the Shapley value (a fair way of sharing
common resources: see [CEW2011]_) as well as test properties of the game
(monotonicity, superadditivity) are also included.

AUTHORS:

- James Campbell and Vince Knight (06-2014): Original version
"""

#*****************************************************************************
#       Copyright (C) 2014 James Campbell james.campbell@tanti.org.uk
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************
from itertools import permutations, combinations
from sage.combinat.subset import powerset
from sage.rings.integer import Integer
from sage.structure.sage_object import SageObject


class CooperativeGame(SageObject):
    r"""
    An object representing a co-operative game. Primarily used to compute the
    Shapley value, but can also provide other information.

    INPUT:

    - ``characteristic_function`` -- dictionary containing all possible
      sets of players:

      * key - each set must be entered as a tuple.
      * value - a real number representing each set of players contribution

    EXAMPLES:

    The type of game that is currently implemented is referred to as a
    Characteristic function game. This is a game on a set of players
    `\Omega` that is defined by a value function `v : C \to \RR` where
    `C = 2^{\Omega}` is the set of all coalitions of players.
    Let `N := |\Omega|`.
    An example of such a game is shown below:

    .. MATH::

        v(c) = \begin{cases}
        0 &\text{if } c = \emptyset, \\
        6 &\text{if } c = \{1\}, \\
        12 &\text{if } c = \{2\}, \\
        42 &\text{if } c = \{3\}, \\
        12 &\text{if } c = \{1,2\}, \\
        42 &\text{if } c = \{1,3\}, \\
        42 &\text{if } c = \{2,3\}, \\
        42 &\text{if } c = \{1,2,3\}. \\
        \end{cases}

    The function `v` can be thought of as a record of contribution of
    individuals and coalitions of individuals. Of interest, becomes how to
    fairly share the value of the grand coalition (`\Omega`)? This class
    allows for such an answer to be formulated by calculating the Shapley
    value of the game.

    Basic examples of how to implement a co-operative game. These functions
    will be used repeatedly in other examples. ::

        sage: integer_function = {(): 0,
        ....:                     (1,): 6,
        ....:                     (2,): 12,
        ....:                     (3,): 42,
        ....:                     (1, 2,): 12,
        ....:                     (1, 3,): 42,
        ....:                     (2, 3,): 42,
        ....:                     (1, 2, 3,): 42}
        sage: integer_game = CooperativeGame(integer_function)

    We can also use strings instead of numbers. ::

        sage: letter_function = {(): 0,
        ....:                    ('A',): 6,
        ....:                    ('B',): 12,
        ....:                    ('C',): 42,
        ....:                    ('A', 'B',): 12,
        ....:                    ('A', 'C',): 42,
        ....:                    ('B', 'C',): 42,
        ....:                    ('A', 'B', 'C',): 42}
        sage: letter_game = CooperativeGame(letter_function)

    Please note that keys should be tuples. ``'1, 2, 3'`` is not a valid key,
    neither is ``123``. The correct input would be ``(1, 2, 3)``. Similarly,
    for coalitions containing a single element the bracket notation (which
    tells Sage that it is a tuple) must be used. So ``(1)``, ``(1,)`` are
    correct however simply inputting `1` is not.

    Characteristic function games can be of various types.

    A characteristic function game `G = (N, v)` is monotone if it satisfies
    `v(C_2) \geq v(C_1)` for all `C_1 \subseteq C_2`. A characteristic
    function game `G = (N, v)` is superadditive if it satisfies
    `v(C_1 \cup C_2) \geq v(C_1) +  v(C_2)` for all `C_1, C_2 \subseteq 2^{\Omega}` such
    that `C_1 \cap C_2 = \emptyset`.

    We can test if a game is monotonic or superadditive. ::

        sage: letter_game.is_monotone()
        True
        sage: letter_game.is_superadditive()
        False

    Instances have a basic representation that will display basic information
    about the game::

        sage: letter_game
        A 3 player co-operative game

    It can be shown that the "fair" payoff vector, referred to as the
    Shapley value is given by the following formula:

    .. MATH::

        \phi_i(G) = \frac{1}{N!} \sum_{\pi\in\Pi_n} \Delta_{\pi}^G(i),

    where the summation is over the permutations of the players and the
    marginal contributions of a player for a given permutation is given as:

    .. MATH::

        \Delta_{\pi}^G(i) = v\bigl( S_{\pi}(i) \cup \{i\} \bigr)
        - v\bigl( S_{\pi}(i) \bigr)

    where `S_{\pi}(i)` is the set of predecessors of `i` in `\pi`, i.e.
    `S_{\pi}(i) = \{ j \mid \pi(i) > \pi(j) \}` (or the number of inversions
    of the form `(i, j)`).

    This payoff vector is "fair" in that it has a collection of properties
    referred to as: efficiency, symmetry, additivity and Null player.
    Some of these properties are considered in this documentation (and tests
    are implemented in the class) but for a good overview see [CEW2011]_.

    Note ([MSZ2013]_) that an equivalent formula for the Shapley value is given by:

    .. MATH::

        \phi_i(G) = \sum_{S \subseteq \Omega} \sum_{p \in S}
        \frac{(|S|-1)!(N-|S|)!}{N!} \bigl( v(S) - v(S \setminus \{p\}) \bigr)
        = \sum_{S \subseteq \Omega} \sum_{p \in S}
        \frac{1}{|S|\binom{N}{|S|}} \bigl( v(S) - v(S \setminus \{p\}) \bigr).

    This later formulation is implemented in Sage and
    requires `2^N-1` calculations instead of `N!`.

    To compute the Shapley value in Sage is simple::

        sage: letter_game.shapley_value()
        {'A': 2, 'B': 5, 'C': 35}

    The following example implements a (trivial) 10 player characteristic
    function game with `v(c) = |c|` for all `c \in 2^{\Omega}`.

    ::

        sage: def simple_characteristic_function(N):
        ....:     return {tuple(coalition) : len(coalition)
        ....:                   for coalition in subsets(range(N))}
        sage: g = CooperativeGame(simple_characteristic_function(10))
        sage: g.shapley_value()
        {0: 1, 1: 1, 2: 1, 3: 1, 4: 1, 5: 1, 6: 1, 7: 1, 8: 1, 9: 1}

    For very large games it might be worth taking advantage of the particular
    problem structure to calculate the Shapley value and there are also
    various approximation approaches to obtaining the Shapley value of a game
    (see [SWJ2008]_ for one such example). Implementing these would be a
    worthwhile development  For more information about the computational
    complexity of calculating the Shapley value see [XP1994]_.

    We can test 3 basic properties of any payoff vector `\lambda`.
    The Shapley value (described above) is known to be the unique
    payoff vector that satisfies these and 1 other property
    not implemented here (additivity). They are:

    * Efficiency - `\sum_{i=1}^N \lambda_i = v(\Omega)`
      In other words, no value of the total coalition is lost.

    * The nullplayer property - If there exists an `i` such that
      `v(C \cup i) = v(C)` for all `C \in 2^{\Omega}` then, `\lambda_i = 0`.
      In other words: if a player does not contribute to any coalition then
      that player should receive no payoff.

    * Symmetry property - If `v(C \cup i) = v(C \cup j)` for all
      `C \in 2^{\Omega} \setminus \{i,j\}`, then `x_i = x_j`.
      If players contribute symmetrically then they should get the same
      payoff::

        sage: payoff_vector = letter_game.shapley_value()
        sage: letter_game.is_efficient(payoff_vector)
        True
        sage: letter_game.nullplayer(payoff_vector)
        True
        sage: letter_game.is_symmetric(payoff_vector)
        True

    Any payoff vector can be passed to the game and these properties
    can once again be tested::

        sage: payoff_vector = {'A': 0, 'C': 35, 'B': 3}
        sage: letter_game.is_efficient(payoff_vector)
        False
        sage: letter_game.nullplayer(payoff_vector)
        True
        sage: letter_game.is_symmetric(payoff_vector)
        True

    TESTS:

    Check that the order within a key does not affect other functions::

        sage: letter_function = {(): 0,
        ....:                    ('A',): 6,
        ....:                    ('B',): 12,
        ....:                    ('C',): 42,
        ....:                    ('A', 'B',): 12,
        ....:                    ('C', 'A',): 42,
        ....:                    ('B', 'C',): 42,
        ....:                    ('B', 'A', 'C',): 42}
        sage: letter_game = CooperativeGame(letter_function)
        sage: letter_game.shapley_value()
        {'A': 2, 'B': 5, 'C': 35}
        sage: letter_game.is_monotone()
        True
        sage: letter_game.is_superadditive()
        False
        sage: letter_game.is_efficient({'A': 2, 'C': 35, 'B': 5})
        True
        sage: letter_game.nullplayer({'A': 2, 'C': 35, 'B': 5})
        True
        sage: letter_game.is_symmetric({'A': 2, 'C': 35, 'B': 5})
        True

    Any payoff vector can be passed to the game and these properties can once
    again be tested. ::

        sage: letter_game.is_efficient({'A': 0, 'C': 35, 'B': 3})
        False
        sage: letter_game.nullplayer({'A': 0, 'C': 35, 'B': 3})
        True
        sage: letter_game.is_symmetric({'A': 0, 'C': 35, 'B': 3})
        True
    """
    def __init__(self, characteristic_function):
        r"""
        Initialize a co-operative game and checks the inputs.

        TESTS:

        An attempt to construct a game from an integer::

            sage: int_game = CooperativeGame(4)
            Traceback (most recent call last):
            ...
            TypeError: characteristic function must be a dictionary

        This test checks that an incorrectly entered singularly tuple will be
        changed into a tuple. In this case ``(1)`` becomes ``(1,)``::

            sage: tuple_function = {(): 0,
            ....:                  (1): 6,
            ....:                  (2,): 12,
            ....:                  (3,): 42,
            ....:                  (1, 2,): 12,
            ....:                  (1, 3,): 42,
            ....:                  (2, 3,): 42,
            ....:                  (1, 2, 3,): 42}
            sage: tuple_game = CooperativeGame(tuple_function)

        This test checks that if a key is not a tuple an error is raised::

            sage: error_function = {(): 0,
            ....:                  (1,): 6,
            ....:                  (2,): 12,
            ....:                  (3,): 42,
            ....:                  12: 12,
            ....:                  (1, 3,): 42,
            ....:                  (2, 3,): 42,
            ....:                  (1, 2, 3,): 42}
            sage: error_game = CooperativeGame(error_function)
            Traceback (most recent call last):
            ...
            TypeError: key must be a tuple

        A test to ensure that the characteristic function is the power
        set of the grand coalition (ie all possible sub-coalitions)::

            sage: incorrect_function = {(): 0,
            ....:                  (1,): 6,
            ....:                  (2,): 12,
            ....:                  (3,): 42,
            ....:                  (1, 2, 3,): 42}
            sage: incorrect_game = CooperativeGame(incorrect_function)
            Traceback (most recent call last):
            ...
            ValueError: characteristic function must be the power set
        """
        if not isinstance(characteristic_function, dict):
            raise TypeError("characteristic function must be a dictionary")

        self.ch_f = characteristic_function
        for key in list(self.ch_f):
            if len(str(key)) == 1 and not isinstance(key, tuple):
                self.ch_f[(key,)] = self.ch_f.pop(key)
            elif not isinstance(key, tuple):
                raise TypeError("key must be a tuple")

        for key in list(self.ch_f):
            sortedkey = tuple(sorted(key))
            self.ch_f[sortedkey] = self.ch_f.pop(key)

        self.player_list = max(characteristic_function, key=len)
        for coalition in powerset(self.player_list):
            if tuple(sorted(coalition)) not in self.ch_f:
                raise ValueError("characteristic function must be the power set")

        self.number_players = len(self.player_list)

    def shapley_value(self):
        r"""
        Return the Shapley value for ``self``.

        The Shapley value is the "fair" payoff vector and
        is computed by the following formula:

        .. MATH::

            \phi_i(G) = \sum_{S \subseteq \Omega} \sum_{p \in S}
            \frac{1}{|S|\binom{N}{|S|}}
            \bigl( v(S) - v(S \setminus \{p\}) \bigr).

        EXAMPLES:

        A typical example of computing the Shapley value::

            sage: integer_function = {(): 0,
            ....:                  (1,): 6,
            ....:                  (2,): 12,
            ....:                  (3,): 42,
            ....:                  (1, 2,): 12,
            ....:                  (1, 3,): 42,
            ....:                  (2, 3,): 42,
            ....:                  (1, 2, 3,): 42}
            sage: integer_game = CooperativeGame(integer_function)
            sage: integer_game.player_list
            (1, 2, 3)
            sage: integer_game.shapley_value()
            {1: 2, 2: 5, 3: 35}

        A longer example of the Shapley value::

            sage: long_function = {(): 0,
            ....:                  (1,): 0,
            ....:                  (2,): 0,
            ....:                  (3,): 0,
            ....:                  (4,): 0,
            ....:                  (1, 2): 0,
            ....:                  (1, 3): 0,
            ....:                  (1, 4): 0,
            ....:                  (2, 3): 0,
            ....:                  (2, 4): 0,
            ....:                  (3, 4): 0,
            ....:                  (1, 2, 3): 0,
            ....:                  (1, 2, 4): 45,
            ....:                  (1, 3, 4): 40,
            ....:                  (2, 3, 4): 0,
            ....:                  (1, 2, 3, 4): 65}
            sage: long_game = CooperativeGame(long_function)
            sage: long_game.shapley_value()
            {1: 70/3, 2: 10, 3: 25/3, 4: 70/3}
        """
        payoff_vector = {}
        n = Integer(len(self.player_list))
        for player in self.player_list:
            weighted_contribution = 0
            for coalition in powerset(self.player_list):
                if coalition:  # If non-empty
                    k = Integer(len(coalition))
                    weight = 1 / (n.binomial(k) * k)
                    t = tuple(p for p in coalition if p != player)
                    weighted_contribution += weight * (self.ch_f[tuple(coalition)]
                                                       - self.ch_f[t])
            payoff_vector[player] = weighted_contribution

        return payoff_vector

    def is_monotone(self):
        r"""
        Return ``True`` if ``self`` is monotonic.

        A game `G = (N, v)` is monotonic if it satisfies
        `v(C_2) \geq v(C_1)` for all `C_1 \subseteq C_2`.

        EXAMPLES:

        A simple game that is monotone::

            sage: integer_function = {(): 0,
            ....:                  (1,): 6,
            ....:                  (2,): 12,
            ....:                  (3,): 42,
            ....:                  (1, 2,): 12,
            ....:                  (1, 3,): 42,
            ....:                  (2, 3,): 42,
            ....:                  (1, 2, 3,): 42}
            sage: integer_game = CooperativeGame(integer_function)
            sage: integer_game.is_monotone()
            True

        An example when the game is not monotone::

            sage: integer_function = {(): 0,
            ....:                  (1,): 6,
            ....:                  (2,): 12,
            ....:                  (3,): 42,
            ....:                  (1, 2,): 10,
            ....:                  (1, 3,): 42,
            ....:                  (2, 3,): 42,
            ....:                  (1, 2, 3,): 42}
            sage: integer_game = CooperativeGame(integer_function)
            sage: integer_game.is_monotone()
            False

        An example on a longer game::

            sage: long_function = {(): 0,
            ....:                  (1,): 0,
            ....:                  (2,): 0,
            ....:                  (3,): 0,
            ....:                  (4,): 0,
            ....:                  (1, 2): 0,
            ....:                  (1, 3): 0,
            ....:                  (1, 4): 0,
            ....:                  (2, 3): 0,
            ....:                  (2, 4): 0,
            ....:                  (3, 4): 0,
            ....:                  (1, 2, 3): 0,
            ....:                  (1, 2, 4): 45,
            ....:                  (1, 3, 4): 40,
            ....:                  (2, 3, 4): 0,
            ....:                  (1, 2, 3, 4): 65}
            sage: long_game = CooperativeGame(long_function)
            sage: long_game.is_monotone()
            True
        """
        return not any(set(p1) <= set(p2) and self.ch_f[p1] > self.ch_f[p2]
                       for p1, p2 in permutations(self.ch_f.keys(), 2))

    def is_superadditive(self):
        r"""
        Return ``True`` if ``self`` is superadditive.

        A characteristic function game `G = (N, v)` is superadditive
        if it satisfies `v(C_1 \cup C_2) \geq v(C_1) +  v(C_2)` for
        all `C_1, C_2 \subseteq 2^{\Omega}` such that `C_1 \cap C_2
        = \emptyset`.

        EXAMPLES:

        An example that is not superadditive::

            sage: integer_function = {(): 0,
            ....:                  (1,): 6,
            ....:                  (2,): 12,
            ....:                  (3,): 42,
            ....:                  (1, 2,): 12,
            ....:                  (1, 3,): 42,
            ....:                  (2, 3,): 42,
            ....:                  (1, 2, 3,): 42}
            sage: integer_game = CooperativeGame(integer_function)
            sage: integer_game.is_superadditive()
            False

        An example that is superadditive::

            sage: A_function = {(): 0,
            ....:               (1,): 6,
            ....:               (2,): 12,
            ....:               (3,): 42,
            ....:               (1, 2,): 18,
            ....:               (1, 3,): 48,
            ....:               (2, 3,): 55,
            ....:               (1, 2, 3,): 80}
            sage: A_game = CooperativeGame(A_function)
            sage: A_game.is_superadditive()
            True

        An example with a longer game that is superadditive::

            sage: long_function = {(): 0,
            ....:                  (1,): 0,
            ....:                  (2,): 0,
            ....:                  (3,): 0,
            ....:                  (4,): 0,
            ....:                  (1, 2): 0,
            ....:                  (1, 3): 0,
            ....:                  (1, 4): 0,
            ....:                  (2, 3): 0,
            ....:                  (2, 4): 0,
            ....:                  (3, 4): 0,
            ....:                  (1, 2, 3): 0,
            ....:                  (1, 2, 4): 45,
            ....:                  (1, 3, 4): 40,
            ....:                  (2, 3, 4): 0,
            ....:                  (1, 2, 3, 4): 65}
            sage: long_game = CooperativeGame(long_function)
            sage: long_game.is_superadditive()
            True

        An example with a longer game that is not::

            sage: long_function = {(): 0,
            ....:                  (1,): 0,
            ....:                  (2,): 0,
            ....:                  (3,): 55,
            ....:                  (4,): 0,
            ....:                  (1, 2): 0,
            ....:                  (1, 3): 0,
            ....:                  (1, 4): 0,
            ....:                  (2, 3): 0,
            ....:                  (2, 4): 0,
            ....:                  (3, 4): 0,
            ....:                  (1, 2, 3): 0,
            ....:                  (1, 2, 4): 45,
            ....:                  (1, 3, 4): 40,
            ....:                  (2, 3, 4): 0,
            ....:                  (1, 2, 3, 4): 85}
            sage: long_game = CooperativeGame(long_function)
            sage: long_game.is_superadditive()
            False
        """
        sets = self.ch_f.keys()
        for p1, p2 in combinations(sets, 2):
            if not (set(p1) & set(p2)):
                union = tuple(sorted(set(p1) | set(p2)))
                if self.ch_f[union] < self.ch_f[p1] + self.ch_f[p2]:
                    return False
        return True

    def _repr_(self):
        r"""
        Return a concise description of ``self``.

        EXAMPLES::

            sage: letter_function = {(): 0,
            ....:                    ('A',): 6,
            ....:                    ('B',): 12,
            ....:                    ('C',): 42,
            ....:                    ('A', 'B',): 12,
            ....:                    ('A', 'C',): 42,
            ....:                    ('B', 'C',): 42,
            ....:                    ('A', 'B', 'C',): 42}
            sage: letter_game = CooperativeGame(letter_function)
            sage: letter_game
            A 3 player co-operative game
        """
        return "A {} player co-operative game".format(self.number_players)

    def _latex_(self):
        r"""
        Return the LaTeX code representing the characteristic function.

        EXAMPLES::

            sage: letter_function = {(): 0,
            ....:                    ('A',): 6,
            ....:                    ('B',): 12,
            ....:                    ('C',): 42,
            ....:                    ('A', 'B',): 12,
            ....:                    ('A', 'C',): 42,
            ....:                    ('B', 'C',): 42,
            ....:                    ('A', 'B', 'C',): 42}
            sage: letter_game = CooperativeGame(letter_function)
            sage: latex(letter_game)
            v(c) = \begin{cases}
            0, & \text{if } c = \emptyset \\
            6, & \text{if } c = \{A\} \\
            12, & \text{if } c = \{B\} \\
            42, & \text{if } c = \{C\} \\
            12, & \text{if } c = \{A, B\} \\
            42, & \text{if } c = \{A, C\} \\
            42, & \text{if } c = \{B, C\} \\
            42, & \text{if } c = \{A, B, C\} \\
            \end{cases}
        """
        cf = self.ch_f
        output = "v(c) = \\begin{cases}\n"
        for key, val in sorted(cf.items(), key=lambda kv: (len(kv[0]), kv[0])):
            if not key:  # == ()
                coalition = "\\emptyset"
            else:
                coalition = "\\{" + ", ".join(str(player) for player in key) + "\\}"
            output += "{}, & \\text{{if }} c = {} \\\\\n".format(val, coalition)
        output += "\\end{cases}"
        return output

    def is_efficient(self, payoff_vector):
        r"""
        Return ``True`` if ``payoff_vector`` is efficient.

        A payoff vector `v` is efficient if
        `\sum_{i=1}^N \lambda_i = v(\Omega)`;
        in other words, no value of the total coalition is lost.

        INPUT:

        - ``payoff_vector`` -- dictionary where the key is the player
          and the value is their payoff

        EXAMPLES:

        An efficient payoff vector::

            sage: letter_function = {(): 0,
            ....:                    ('A',): 6,
            ....:                    ('B',): 12,
            ....:                    ('C',): 42,
            ....:                    ('A', 'B',): 12,
            ....:                    ('A', 'C',): 42,
            ....:                    ('B', 'C',): 42,
            ....:                    ('A', 'B', 'C',): 42}
            sage: letter_game = CooperativeGame(letter_function)
            sage: letter_game.is_efficient({'A': 14, 'B': 14, 'C': 14})
            True

            sage: letter_function = {(): 0,
            ....:                    ('A',): 6,
            ....:                    ('B',): 12,
            ....:                    ('C',): 42,
            ....:                    ('A', 'B',): 12,
            ....:                    ('A', 'C',): 42,
            ....:                    ('B', 'C',): 42,
            ....:                    ('A', 'B', 'C',): 42}
            sage: letter_game = CooperativeGame(letter_function)
            sage: letter_game.is_efficient({'A': 10, 'B': 14, 'C': 14})
            False

        A longer example::

            sage: long_function = {(): 0,
            ....:                  (1,): 0,
            ....:                  (2,): 0,
            ....:                  (3,): 0,
            ....:                  (4,): 0,
            ....:                  (1, 2): 0,
            ....:                  (1, 3): 0,
            ....:                  (1, 4): 0,
            ....:                  (2, 3): 0,
            ....:                  (2, 4): 0,
            ....:                  (3, 4): 0,
            ....:                  (1, 2, 3): 0,
            ....:                  (1, 2, 4): 45,
            ....:                  (1, 3, 4): 40,
            ....:                  (2, 3, 4): 0,
            ....:                  (1, 2, 3, 4): 65}
            sage: long_game = CooperativeGame(long_function)
            sage: long_game.is_efficient({1: 20, 2: 20, 3: 5, 4: 20})
            True
        """
        pl = tuple(sorted(self.player_list))
        return sum(payoff_vector.values()) == self.ch_f[pl]

    def nullplayer(self, payoff_vector):
        r"""
        Return ``True`` if ``payoff_vector`` possesses the nullplayer
        property.

        A payoff vector `v` has the nullplayer property if there exists
        an `i` such that `v(C \cup i) = v(C)` for all `C \in 2^{\Omega}`
        then, `\lambda_i = 0`. In other words: if a player does not
        contribute to any coalition then that player should receive no payoff.

        INPUT:

        - ``payoff_vector`` -- dictionary where the key is the player
          and the value is their payoff

        EXAMPLES:

        A payoff vector that returns ``True``::

            sage: letter_function = {(): 0,
            ....:                    ('A',): 0,
            ....:                    ('B',): 12,
            ....:                    ('C',): 42,
            ....:                    ('A', 'B',): 12,
            ....:                    ('A', 'C',): 42,
            ....:                    ('B', 'C',): 42,
            ....:                    ('A', 'B', 'C',): 42}
            sage: letter_game = CooperativeGame(letter_function)
            sage: letter_game.nullplayer({'A': 0, 'B': 14, 'C': 14})
            True

        A payoff vector that returns ``False``::

            sage: A_function = {(): 0,
            ....:               (1,): 0,
            ....:               (2,): 12,
            ....:               (3,): 42,
            ....:               (1, 2,): 12,
            ....:               (1, 3,): 42,
            ....:               (2, 3,): 55,
            ....:               (1, 2, 3,): 55}
            sage: A_game = CooperativeGame(A_function)
            sage: A_game.nullplayer({1: 10, 2: 10, 3: 25})
            False

        A longer example for nullplayer::

            sage: long_function = {(): 0,
            ....:                  (1,): 0,
            ....:                  (2,): 0,
            ....:                  (3,): 0,
            ....:                  (4,): 0,
            ....:                  (1, 2): 0,
            ....:                  (1, 3): 0,
            ....:                  (1, 4): 0,
            ....:                  (2, 3): 0,
            ....:                  (2, 4): 0,
            ....:                  (3, 4): 0,
            ....:                  (1, 2, 3): 0,
            ....:                  (1, 2, 4): 45,
            ....:                  (1, 3, 4): 40,
            ....:                  (2, 3, 4): 0,
            ....:                  (1, 2, 3, 4): 65}
            sage: long_game = CooperativeGame(long_function)
            sage: long_game.nullplayer({1: 20, 2: 20, 3: 5, 4: 20})
            True

        TESTS:

        Checks that the function is going through all players::

            sage: A_function = {(): 0,
            ....:               (1,): 42,
            ....:               (2,): 12,
            ....:               (3,): 0,
            ....:               (1, 2,): 55,
            ....:               (1, 3,): 42,
            ....:               (2, 3,): 12,
            ....:               (1, 2, 3,): 55}
            sage: A_game = CooperativeGame(A_function)
            sage: A_game.nullplayer({1: 10, 2: 10, 3: 25})
            False
        """
        for player in self.player_list:
            results = []
            for coalit in self.ch_f:
                if player in coalit:
                    t = tuple(sorted(set(coalit) - {player}))
                    results.append(self.ch_f[coalit] == self.ch_f[t])
            if all(results) and payoff_vector[player] != 0:
                return False
        return True

    def is_symmetric(self, payoff_vector):
        r"""
        Return ``True`` if ``payoff_vector`` possesses the symmetry property.

        A payoff vector possesses the symmetry property if
        `v(C \cup i) = v(C \cup j)` for all
        `C \in 2^{\Omega} \setminus \{i,j\}`, then `x_i = x_j`.

        INPUT:

        - ``payoff_vector`` -- dictionary where the key is the player
          and the value is their payoff

        EXAMPLES:

        A payoff vector that has the symmetry property::

            sage: letter_function = {(): 0,
            ....:                    ('A',): 6,
            ....:                    ('B',): 12,
            ....:                    ('C',): 42,
            ....:                    ('A', 'B',): 12,
            ....:                    ('A', 'C',): 42,
            ....:                    ('B', 'C',): 42,
            ....:                    ('A', 'B', 'C',): 42}
            sage: letter_game = CooperativeGame(letter_function)
            sage: letter_game.is_symmetric({'A': 5, 'B': 14, 'C': 20})
            True

        A payoff vector that returns ``False``::

            sage: integer_function = {(): 0,
            ....:                     (1,): 12,
            ....:                     (2,): 12,
            ....:                     (3,): 42,
            ....:                     (1, 2,): 12,
            ....:                     (1, 3,): 42,
            ....:                     (2, 3,): 42,
            ....:                     (1, 2, 3,): 42}
            sage: integer_game = CooperativeGame(integer_function)
            sage: integer_game.is_symmetric({1: 2, 2: 5, 3: 35})
            False

        A longer example for symmetry::

            sage: long_function = {(): 0,
            ....:                  (1,): 0,
            ....:                  (2,): 0,
            ....:                  (3,): 0,
            ....:                  (4,): 0,
            ....:                  (1, 2): 0,
            ....:                  (1, 3): 0,
            ....:                  (1, 4): 0,
            ....:                  (2, 3): 0,
            ....:                  (2, 4): 0,
            ....:                  (3, 4): 0,
            ....:                  (1, 2, 3): 0,
            ....:                  (1, 2, 4): 45,
            ....:                  (1, 3, 4): 40,
            ....:                  (2, 3, 4): 0,
            ....:                  (1, 2, 3, 4): 65}
            sage: long_game = CooperativeGame(long_function)
            sage: long_game.is_symmetric({1: 20, 2: 20, 3: 5, 4: 20})
            True
        """
        sets = self.ch_f.keys()
        element = [i for i in sets if len(i) == 1]
        for c1, c2 in combinations(element, 2):
            results = []
            for m in sets:
                junion = tuple(sorted(set(c1) | set(m)))
                kunion = tuple(sorted(set(c2) | set(m)))
                results.append(self.ch_f[junion] == self.ch_f[kunion])
            if all(results) and payoff_vector[c1[0]] != payoff_vector[c2[0]]:
                return False
        return True



# sage_setup: distribution = sagemath-polyhedra
r"""
Ordered Hyperplane Arrangements

The :class:`HyperplaneArrangements` orders the hyperplanes in a arrangement
independently of the way the hyperplanes are introduced. The class
:class:`OrderedHyperplaneArrangements` fixes an order specified by
the user. This can be needed for certain properties, e.g., fundamental group with
information about meridians, braid monodromy with information about the strands;
in the future, it may be useful for combinatorial properties.
There are no other differences with usual hyperplane arrangements.

An ordered arrangement is an arrangement where the hyperplanes are sorted
by the user::

    sage: H0.<t0, t1, t2> = HyperplaneArrangements(QQ)
    sage: H0(t0 - t1, t1 - t2, t0 - t2)
    Arrangement <t1 - t2 | t0 - t1 | t0 - t2>
    sage: H.<t0, t1, t2> = OrderedHyperplaneArrangements(QQ)
    sage: H(t0 - t1, t1 - t2, t0 - t2)
    Arrangement <t0 - t1 | t1 - t2 | t0 - t2>

Some methods are adapted, e.g., :meth:`~sage.geometry.hyperplane_arrangement.arrangement.HyperplaneArrangementElement.hyperplanes`,
and some new ones are created, regarding
hyperplane sections and fundamental groups::

    sage: H.<x,y> = HyperplaneArrangements(QQ)
    sage: H1.<x,y> = OrderedHyperplaneArrangements(QQ)
    sage: A1 = H1(x, y); A = H(A1)
    sage: A.hyperplanes()
    (Hyperplane 0*x + y + 0, Hyperplane x + 0*y + 0)
    sage: A1.hyperplanes()
    (Hyperplane x + 0*y + 0, Hyperplane 0*x + y + 0)

We see the differences in :meth:`~sage.geometry.hyperplane_arrangement.arrangement.HyperplaneArrangementElement.union`::

    sage: H.<x,y> = HyperplaneArrangements(QQ)
    sage: H1.<x,y> = OrderedHyperplaneArrangements(QQ)
    sage: A = H([1,2,3], [0,1,1], [0,1,-1], [1,-1,0], [1,1,0])
    sage: B = H([1,1,1], [1,-1,1], [1,0,-1])
    sage: C = A.union(B)
    sage: A1 = H1(A); B1 = H1(B); C1 = A1.union(B1)
    sage: [C1.hyperplanes().index(h) for h in C.hyperplanes()]
    [0, 5, 6, 1, 2, 3, 7, 4]

Also in meth:`~sage.geometry.hyperplane_arrangement.arrangement.HyperplaneArrangementElement.cone`::

    sage: # needs sage.combinat
    sage: a.<x,y,z> = hyperplane_arrangements.semiorder(3)
    sage: H.<x,y,z> = OrderedHyperplaneArrangements(QQ)
    sage: a1 = H(a)
    sage: b = a.cone(); b1 = a1.cone()
    sage: [b1.hyperplanes().index(h) for h in b.hyperplanes()]
    [0, 2, 4, 6, 1, 3, 5]

And in :meth:`~sage.geometry.hyperplane_arrangement.arrangement.HyperplaneArrangementElement.restriction`::

    sage: # needs sage.graphs
    sage: A.<u, x, y, z> = hyperplane_arrangements.braid(4)
    sage: L.<u, x, y, z> = OrderedHyperplaneArrangements(QQ)
    sage: A1 = L(A)
    sage: H = A[0]; H
    Hyperplane 0*u + 0*x + y - z + 0
    sage: A.restriction(H)
    Arrangement <x - z | u - x | u - z>
    sage: A1.restriction(H)
    Arrangement <x - z | u - x | u - z>
    sage: A1.restriction(H, repetitions=True)
    Arrangement of 5 hyperplanes of dimension 3 and rank 2

AUTHORS:

- Enrique Artal (2023-12): initial version

This module adds some features to the *unordered* one for some
properties which depend on the order.
"""

# *****************************************************************************
#       Copyright (C) 2023 Enrique Artal <artal@unizar.es>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
# *****************************************************************************

from sage.geometry.hyperplane_arrangement.arrangement import HyperplaneArrangementElement
from sage.geometry.hyperplane_arrangement.arrangement import HyperplaneArrangements
from sage.geometry.hyperplane_arrangement.hyperplane import Hyperplane
from sage.matrix.constructor import matrix, vector
from sage.misc.lazy_import import lazy_import
from sage.misc.misc_c import prod
from sage.rings.integer_ring import ZZ

lazy_import('sage.groups.free_group', 'FreeGroup')
lazy_import('sage.rings.qqbar', 'QQbar')
lazy_import('sage.schemes.curves.plane_curve_arrangement', 'AffinePlaneCurveArrangements')
lazy_import('sage.schemes.curves.plane_curve_arrangement', 'ProjectivePlaneCurveArrangements')


class OrderedHyperplaneArrangementElement(HyperplaneArrangementElement):
    """
    An ordered hyperplane arrangement.

    .. WARNING::

        You should never create
        :class:`OrderedHyperplaneArrangementElement` instances directly,
        always use the parent.
    """
    def __init__(self, parent, hyperplanes, check=True, backend=None):
        """
        Construct an ordered hyperplane arrangement.

        INPUT:

        - ``parent`` -- the parent :class:`OrderedHyperplaneArrangements`

        - ``hyperplanes`` -- tuple of hyperplanes

        - ``check`` -- boolean (default: ``True``); whether
          to check input

        - ``backend`` -- string (default: ``None``); the backend to
          use for the related polyhedral objects

        EXAMPLES::

            sage: H.<x,y> = OrderedHyperplaneArrangements(QQ)
            sage: elt = H(x, y); elt
            Arrangement <x | y>
            sage: TestSuite(elt).run()
        """
        super().__init__(parent, hyperplanes, check=check, backend=backend)
        self._affine_fundamental_group = None
        self._affine_meridians = None
        self._projective_fundamental_group = None
        self._projective_meridians = None

    def hyperplane_section(self, proj=True):
        r"""
        Compute a generic hyperplane section of ``self``.

        INPUT:

        - ``proj`` -- (default: ``True``) if the
          ambient space is affine or projective

        OUTPUT:

        An arrangement `\mathcal{A}` obtained by intersecting with a
        generic hyperplane

        EXAMPLES::

            sage: L.<x, y, z> = OrderedHyperplaneArrangements(QQ)
            sage: L(x, y - 1, z).hyperplane_section()
            Traceback (most recent call last):
            ...
            TypeError: the arrangement is not projective

            sage: # needs sage.graphs
            sage: A0.<u,x,y,z> = hyperplane_arrangements.braid(4); A0
            Arrangement of 6 hyperplanes of dimension 4 and rank 3
            sage: L.<u,x,y,z> = OrderedHyperplaneArrangements(QQ)
            sage: A = L(A0)
            sage: M = A.matroid()
            sage: A1 = A.hyperplane_section()
            sage: A1
            Arrangement of 6 hyperplanes of dimension 3 and rank 3
            sage: M1 = A1.matroid()
            sage: A2 = A1.hyperplane_section(); A2
            Arrangement of 6 hyperplanes of dimension 2 and rank 2
            sage: M2 = A2.matroid()
            sage: T1 = M1.truncation()
            sage: T1.is_isomorphic(M2)
            True
            sage: T1.isomorphism(M2)
            {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5}

            sage: # needs sage.combinat
            sage: a0 = hyperplane_arrangements.semiorder(3); a0
            Arrangement of 6 hyperplanes of dimension 3 and rank 2
            sage: L.<t0, t1, t2> = OrderedHyperplaneArrangements(QQ)
            sage: a = L(a0)
            sage: ca = a.cone()
            sage: m = ca.matroid()
            sage: a1 = a.hyperplane_section(proj=False)
            sage: a1
            Arrangement of 6 hyperplanes of dimension 2 and rank 2
            sage: ca1 = a1.cone()
            sage: m1 = ca1.matroid()
            sage: m.isomorphism(m1)
            {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6}
            sage: p0 = hyperplane_arrangements.Shi(4)
            sage: L.<t0, t1, t2, t3> = OrderedHyperplaneArrangements(QQ)
            sage: p = L(p0)
            sage: a = p.hyperplane_section(proj=False); a
            Arrangement of 12 hyperplanes of dimension 3 and rank 3
            sage: ca = a.cone()
            sage: m = ca.matroid().truncation()
            sage: a1 = a.hyperplane_section(proj=False); a1
            Arrangement of 12 hyperplanes of dimension 2 and rank 2
            sage: ca1 = a1.cone()
            sage: m1 = ca1.matroid()
            sage: m1.is_isomorphism(m, {j: j for j in range(13)})
            True
        """
        if proj and not self.is_linear():
            raise TypeError('the arrangement is not projective')
        n0 = self.dimension()
        if not proj:
            H = self.cone()
            H1 = H.hyperplane_section()
            mat = matrix(h.coefficients()[1:] for h in H1)
            m = mat.nrows()
            for j in range(mat.ncols()):
                if mat[m - 1, j] != 0:
                    mat.swap_columns(0, j)
                    break
            for j in range(1, mat.ncols()):
                mat.add_multiple_of_column(j, 0, -mat[m - 1, j] / mat[m - 1, 0])
            vrs = H1.parent().variable_names()[1:]
            A1 = OrderedHyperplaneArrangements(self.base_ring(), names=vrs)
            mat_rows = mat.rows()[:-1]
            H1b = A1(mat_rows)
            return H1b
        P = self.intersection_poset(element_label='subspace')
        center = P.maximal_elements()[0].linear_part()
        n1 = center.dimension()
        U = []
        for p in P:
            if p.dimension() == n1 + 1:
                B = [u for u in p.linear_part().basis() if u not in center]
                U.append(B[0])
        # U = [p.linear_part().basis()[0] for p in P if p.dimension() == n1 + 1]
        U0 = sum(U)
        # We look for a linear hyperplane with integer coefficients
        # defining a transversal hyperplane
        for v in ZZ**n0:
            v1 = v + U0
            if 0 not in [w * v1 for w in U]:
                break
        h0 = self.parent()((0,) + tuple(v1))
        H1 = self.add_hyperplane(h0)
        return H1.restriction(h0)

    def affine_fundamental_group(self):
        r"""
        Return the fundamental group of the complement of an affine
        hyperplane arrangement in `\CC^n` whose equations have
        coefficients in a subfield of `\QQbar`.

        OUTPUT: a finitely presented fundamental group

        .. NOTE::

           This functionality requires the ``sirocco`` package to be installed.

        EXAMPLES::

            sage: # needs sirocco
            sage: A.<x, y> = OrderedHyperplaneArrangements(QQ)
            sage: L = [y + x, y + x - 1]
            sage: H = A(L)
            sage: H.affine_fundamental_group()
            Finitely presented group < x0, x1 |  >
            sage: L = [x, y, x + 1, y + 1, x - y]
            sage: A(L).affine_fundamental_group()
            Finitely presented group
            < x0, x1, x2, x3, x4 | x4*x0*x4^-1*x0^-1,
                                   x0*x2*x3*x2^-1*x0^-1*x3^-1,
                                   x1*x2*x4*x2^-1*x1^-1*x4^-1,
                                   x2*x3*x0*x2^-1*x0^-1*x3^-1,
                                   x2*x4*x1*x2^-1*x1^-1*x4^-1,
                                   x4*x1*x4^-1*x3^-1*x2^-1*x1^-1*x2*x3 >
            sage: H = A(x, y, x + y)
            sage: H.affine_fundamental_group()
            Finitely presented group
            < x0, x1, x2 | x0*x1*x2*x1^-1*x0^-1*x2^-1, x1*x2*x0*x1^-1*x0^-1*x2^-1 >
            sage: H.affine_fundamental_group()  # repeat to use the attribute
            Finitely presented group
            < x0, x1, x2 | x0*x1*x2*x1^-1*x0^-1*x2^-1, x1*x2*x0*x1^-1*x0^-1*x2^-1 >
            sage: T.<t> = QQ[]
            sage: K.<a> = NumberField(t^3 + t + 1)
            sage: L.<x, y> = OrderedHyperplaneArrangements(K)
            sage: H = L(a*x + y -1, x + a*y + 1, x - 1, y - 1)
            sage: H.affine_fundamental_group()
            Traceback (most recent call last):
            ...
            TypeError: the base field is not in QQbar
            sage: L.<t> = OrderedHyperplaneArrangements(QQ)
            sage: L([t - j for j in range(4)]).affine_fundamental_group()
            Finitely presented group < x0, x1, x2, x3 |  >
            sage: L.<x, y, z> = OrderedHyperplaneArrangements(QQ)
            sage: L(L.gens() + (x + y + z + 1,)).affine_fundamental_group().sorted_presentation()
            Finitely presented group
            < x0, x1, x2, x3 | x3^-1*x2^-1*x3*x2, x3^-1*x1^-1*x3*x1,
                               x3^-1*x0^-1*x3*x0, x2^-1*x1^-1*x2*x1,
                               x2^-1*x0^-1*x2*x0, x1^-1*x0^-1*x1*x0 >
            sage: A = OrderedHyperplaneArrangements(QQ, names=())
            sage: H = A(); H
            Empty hyperplane arrangement of dimension 0
            sage: H.affine_fundamental_group()
            Finitely presented group <  |  >
        """
        K = self.base_ring()
        if not K.is_subring(QQbar):
            raise TypeError('the base field is not in QQbar')
        if self._affine_fundamental_group:
            return self._affine_fundamental_group
        n = self.dimension()
        r = len(self)
        if n == 0:
            return FreeGroup(0) / []
        if n == 1:
            G = FreeGroup(r) / []
            dic = {j: G.gen(j) for j in range(r)}
            dic[r] = [prod(G.gens()) ** -1]
            self._affine_fundamental_group = G
            self._affine_meridians = dic
            return G
        if n == 2:
            S = self.parent().ambient_space().symmetric_space()
            coord = vector((1,) + S.gens())
            Af = AffinePlaneCurveArrangements(K, names=self.parent().variable_names())
            L = Af([vector(line.coefficients()) * coord for line in self])
            G = L.fundamental_group()
            self._affine_fundamental_group = G
            self._affine_meridians = L._meridians_simpl_vertical
            return G
        H1 = self.hyperplane_section(proj=False)
        G = H1.affine_fundamental_group()
        self._affine_fundamental_group = G
        self._affine_meridians = H1._affine_meridians
        return G

    def affine_meridians(self):
        r"""
        Return the meridians of each hyperplane (including the one at infinity).

        OUTPUT: a dictionary

        .. NOTE::

           This functionality requires the ``sirocco`` package to be installed.

        EXAMPLES::

            sage: # needs sirocco
            sage: A.<x, y> = OrderedHyperplaneArrangements(QQ)
            sage: L = [y + x, y + x - 1]
            sage: H = A(L)
            sage: g = H.affine_fundamental_group()
            sage: g
            Finitely presented group < x0, x1 |  >
            sage: H.affine_meridians()
            {0: [x0], 1: [x1], 2: [x1^-1*x0^-1]}
            sage: H1 = H.add_hyperplane(y - x)
            sage: H1.affine_meridians()
            {0: [x0], 1: [x1], 2: [x2], 3: [x2^-1*x1^-1*x0^-1]}
        """
        if self._affine_meridians is None:
            self.affine_fundamental_group()
        return dict(self._affine_meridians)

    def projective_fundamental_group(self):
        r"""
        Return the fundamental group of the complement of a projective
        hyperplane arrangement.

        OUTPUT:

        The finitely presented group of the complement
        in the projective space whose equations have
        coefficients in a subfield of `\QQbar`.

        .. NOTE::

           This functionality requires the ``sirocco`` package to be installed.

        EXAMPLES::

            sage: # needs sirocco
            sage: A.<x, y> = OrderedHyperplaneArrangements(QQ)
            sage: H = A(x, y, x + y)
            sage: H.projective_fundamental_group()
            Finitely presented group < x0, x1 |  >

            sage: # needs sirocco sage.graphs
            sage: A3.<x, y, z> = OrderedHyperplaneArrangements(QQ)
            sage: H = A3(hyperplane_arrangements.braid(4).essentialization())
            sage: G3 = H.projective_fundamental_group(); G3.sorted_presentation()
            Finitely presented group
            < x0, x1, x2, x3, x4 | x4^-1*x3^-1*x2^-1*x3*x4*x0*x2*x0^-1,
                                   x4^-1*x2^-1*x4*x2, x4^-1*x1^-1*x0^-1*x1*x4*x0,
                                   x4^-1*x1^-1*x0^-1*x4*x0*x1,
                                   x4^-1*x1^-1*x3*x0*x1*x3^-1*x2^-1*x4*x0^-1*x2,
                                   x3^-1*x2^-1*x1^-1*x0^-1*x3*x0*x1*x2,
                                   x3^-1*x1^-1*x3*x1 >
            sage: G3.abelian_invariants()
            (0, 0, 0, 0, 0)
            sage: A4.<t1, t2, t3, t4> = OrderedHyperplaneArrangements(QQ)
            sage: H = A4(hyperplane_arrangements.braid(4))
            sage: G4 = H.projective_fundamental_group(); G4.sorted_presentation()
            Finitely presented group
            < x0, x1, x2, x3, x4 | x4^-1*x3^-1*x2^-1*x3*x4*x0*x2*x0^-1,
                                   x4^-1*x2^-1*x4*x2, x4^-1*x1^-1*x0^-1*x1*x4*x0,
                                   x4^-1*x1^-1*x0^-1*x4*x0*x1,
                                   x4^-1*x1^-1*x3*x0*x1*x3^-1*x2^-1*x4*x0^-1*x2,
                                   x3^-1*x2^-1*x1^-1*x0^-1*x3*x0*x1*x2,
                                   x3^-1*x1^-1*x3*x1 >
            sage: G4.abelian_invariants()
            (0, 0, 0, 0, 0)

            sage: # needs sirocco
            sage: L.<t0, t1, t2, t3, t4> = OrderedHyperplaneArrangements(QQ)
            sage: H = hyperplane_arrangements.coordinate(5)
            sage: H = L(H)
            sage: g = H.projective_fundamental_group()
            sage: g.is_abelian(), g.abelian_invariants()
            (True, (0, 0, 0, 0))
            sage: L(t0, t1, t2, t3, t4, t0 - 1).projective_fundamental_group()
            Traceback (most recent call last):
            ...
            TypeError: the arrangement is not projective
            sage: T.<t> = QQ[]
            sage: K.<a> = NumberField(t^3 + t + 1)
            sage: L.<x, y, z> = OrderedHyperplaneArrangements(K)
            sage: H = L(a*x + y - z, x + a*y + z, x - z, y - z)
            sage: H.projective_fundamental_group()
            Traceback (most recent call last):
            ...
            TypeError: the base field is not in QQbar
            sage: A.<x> = OrderedHyperplaneArrangements(QQ)
            sage: H = A(); H
            Empty hyperplane arrangement of dimension 1
            sage: H.projective_fundamental_group()
            Finitely presented group <  |  >
        """
        K = self.base_ring()
        if not K.is_subring(QQbar):
            raise TypeError('the base field is not in QQbar')
        if not self.is_linear():
            raise TypeError('the arrangement is not projective')
        if self._projective_fundamental_group:
            return self._projective_fundamental_group
        n = self.dimension()
        r = len(self)
        if n == 1:
            return FreeGroup(0) / []
        if n == 2:
            G = FreeGroup(r - 1) / []
            dic = {j: G.gen(j) for j in range(r - 1)}
            dic[r - 1] = [prod(G.gens()) ** -1]
            self._projective_fundamental_group = G
            self._projective_meridians = dic
            return G
        if n == 3:
            S = self.parent().ambient_space().symmetric_space()
            coord = vector(S.gens())
            Proj = ProjectivePlaneCurveArrangements(K,
                                                    names=self.parent().variable_names())
            L = Proj([vector(line.coefficients()[1:]) * coord for line in self])
            G = L.fundamental_group()
            self._projective_fundamental_group = G
            self._projective_meridians = L._meridians_simpl
            return G
        H1 = self.hyperplane_section()
        G = H1.projective_fundamental_group()
        self._projective_fundamental_group = G
        self._projective_meridians = H1._projective_meridians
        return G

    def projective_meridians(self):
        r"""
        Return the meridian of each hyperplane.

        OUTPUT: a dictionary

        .. NOTE::

           This functionality requires the ``sirocco`` package to be installed.

        EXAMPLES::

            sage: # needs sirocco
            sage: A.<x, y> = OrderedHyperplaneArrangements(QQ)
            sage: H = A(x, y, x + y)
            sage: H.projective_meridians()
            {0: x0, 1: x1, 2: [x1^-1*x0^-1]}

            sage: # needs sirocco sage.graphs
            sage: A3.<x, y, z> = OrderedHyperplaneArrangements(QQ)
            sage: H = A3(hyperplane_arrangements.braid(4).essentialization())
            sage: H.projective_meridians()
            {0: [x2^-1*x0^-1*x4^-1*x3^-1*x1^-1],
             1: [x3], 2: [x4], 3: [x1], 4: [x2], 5: [x0]}
            sage: A4.<t1, t2, t3, t4> = OrderedHyperplaneArrangements(QQ)
            sage: H = A4(hyperplane_arrangements.braid(4))
            sage: H.projective_meridians()
            {0: [x2^-1*x0^-1*x4^-1*x3^-1*x1^-1], 1: [x3],
             2: [x4], 3: [x0], 4: [x2], 5: [x1]}

            sage: # needs sirocco
            sage: L.<t0, t1, t2, t3, t4> = OrderedHyperplaneArrangements(QQ)
            sage: H = hyperplane_arrangements.coordinate(5)
            sage: H = L(H)
            sage: H.projective_meridians()
            {0: [x2], 1: [x3], 2: [x0], 3: [x3^-1*x2^-1*x1^-1*x0^-1], 4: [x1]}
        """
        if self._projective_meridians is None:
            self.projective_fundamental_group()
        return dict(self._projective_meridians)


class OrderedHyperplaneArrangements(HyperplaneArrangements):
    """
    Ordered Hyperplane arrangements.

    For more information on hyperplane arrangements, see
    :mod:`sage.geometry.hyperplane_arrangement.arrangement`.

    INPUT:

    - ``base_ring`` -- ring; the base ring

    - ``names`` -- tuple of strings; the variable names

    EXAMPLES::

        sage: H.<x,y> = HyperplaneArrangements(QQ)
        sage: x
        Hyperplane x + 0*y + 0
        sage: x + y
        Hyperplane x + y + 0
        sage: H(x, y, x-1, y-1)
        Arrangement <y - 1 | y | x - 1 | x>
    """
    Element = OrderedHyperplaneArrangementElement

    def _element_constructor_(self, *args, **kwds):
        """
        Construct an element of ``self``.

        INPUT:

        - ``*args`` -- positional arguments, each defining a
          hyperplane; alternatively, a single polytope or a single
          hyperplane arrangement

        - ``signed`` -- boolean (default: ``True``); whether to
          preserve signs of hyperplane equations

        - ``check`` -- boolean (default: ``True``); whether to
          perform argument checking

        EXAMPLES::

            sage: L.<x, y> = OrderedHyperplaneArrangements(QQ)
            sage: L(x)
            Arrangement <x>
            sage: L(x, y)
            Arrangement <x | y>
            sage: L([x, y])
            Arrangement <x | y>
            sage: L([0, 1, 0], [0, 0, 1])
            Arrangement <x | y>
            sage: L([[0, 0, 1], [0, 1, 0]])
            Arrangement <y | x>

            sage: L(polytopes.hypercube(2))
            Arrangement <-x + 1 | -y + 1 | x + 1 | y + 1>

            sage: L(-x, x + y - 1, signed=False)
            Arrangement <x | -x - y + 1>

        TESTS::

            sage: L()
            Empty hyperplane arrangement of dimension 2
            sage: L(0)        # zero is equivalent to no argument, Issue #8648
            Empty hyperplane arrangement of dimension 2
            sage: L(0*x)      # degenerate hyperplane is NOT allowed
            Traceback (most recent call last):
            ...
            ValueError: linear expression must be non-constant to define a hyperplane
            sage: L(0*x, y)   # ditto
            Traceback (most recent call last):
            ...
            ValueError: linear expression must be non-constant to define a hyperplane
        """
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, HyperplaneArrangementElement) and arg.parent() is self:
                # optimization if argument is already a hyperplane arrangement
                return arg
            if arg == 0 and not isinstance(arg, Hyperplane):
                # zero = neutral element under addition = the empty hyperplane arrangement
                args = []
        # process keyword arguments
        not_char2 = (self.base_ring().characteristic() != 2)
        signed = kwds.pop('signed', not_char2)
        check = kwds.pop('check', True)
        backend = kwds.pop('backend', None)
        if kwds:
            raise ValueError('unknown keyword argument')
        # process positional arguments
        AA = self.ambient_space()
        try:
            hyperplanes = [AA(a) for a in args]
        except (TypeError, ValueError, AttributeError):
            if len(args) > 1:
                raise
            arg = args[0]
            if hasattr(arg, 'Hrepresentation'):
                hyperplanes = [AA(h) for h in arg.Hrepresentation()]
            else:
                hyperplanes = [AA(a) for a in arg]
        hyperplanes = [h.primitive(signed) for h in hyperplanes]
        if check:
            if signed and not not_char2:
                raise ValueError('cannot be signed in characteristic 2')
            for h in hyperplanes:
                if h.A() == 0:
                    raise ValueError('linear expression must be non-constant to define a hyperplane')
                if not_char2 and -h in hyperplanes:
                    raise ValueError('arrangement cannot simultaneously have h and -h as hyperplane')
        return self.element_class(self, tuple(hyperplanes), backend=backend)

    def _repr_(self):
        """
        Return a string representation.

        OUTPUT: string

        EXAMPLES::

            sage: L.<x, y> = OrderedHyperplaneArrangements(QQ);  L
            Ordered hyperplane arrangements in 2-dimensional linear space
            over Rational Field with coordinates x, y
        """
        return 'Ordered hyperplane arrangements in {0}'.format(self.ambient_space())



# sage_setup: distribution = sagemath-polyhedra
# sage.doctest: needs sage.geometry.polyhedron sage.graphs
r"""
Subschemes of toric space

AUTHORS:

- David Kohel (2005): initial version.

- William Stein (2005): initial version.

- Andrey Novoseltsev (2010-05-17): subschemes of toric varieties.
"""

# ****************************************************************************
# Copyright (C) 2005 William Stein <wstein@gmail.com>
#
# Distributed under the terms of the GNU General Public License (GPL)
# as published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
# https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.lazy_import import lazy_import
from sage.rings.integer_ring import ZZ
from sage.schemes.generic.algebraic_scheme import AlgebraicScheme_subscheme

lazy_import("sage.calculus.functions", "jacobian")


class AlgebraicScheme_subscheme_toric(AlgebraicScheme_subscheme):
    r"""
    Construct an algebraic subscheme of a toric variety.

    .. WARNING::

        You should not create objects of this class directly. The
        preferred method to construct such subschemes is to use
        :meth:`~ToricVariety_field.subscheme` method of :class:`toric
        varieties <sage.schemes.toric.variety.ToricVariety_field>`.

    INPUT:

    - ``toric_variety`` -- ambient :class:`toric variety
      <ToricVariety_field>`

    - ``polynomials`` -- single polynomial, list, or ideal of defining
      polynomials in the coordinate ring of ``toric_variety``

    OUTPUT: an :class:`algebraic subscheme of a toric variety
    <AlgebraicScheme_subscheme_toric>`.

    TESTS::

        sage: P1xP1 = toric_varieties.P1xP1()
        sage: P1xP1.inject_variables()
        Defining s, t, x, y
        sage: import sage.schemes.toric.toric_subscheme as SCM
        sage: X = SCM.AlgebraicScheme_subscheme_toric(
        ....:       P1xP1, [x*s + y*t, x^3 + y^3])
        sage: X
        Closed subscheme of 2-d CPR-Fano toric variety
        covered by 4 affine patches defined by:
          s*x + t*y,
          x^3 + y^3

    A better way to construct the same scheme as above::

        sage: P1xP1.subscheme([x*s + y*t, x^3 + y^3])
        Closed subscheme of 2-d CPR-Fano toric variety
        covered by 4 affine patches defined by:
          s*x + t*y,
          x^3 + y^3
    """

    # Implementation note: if the toric variety is affine you should
    # construct instances of the derived class
    # AlgebraicScheme_subscheme_affine_toric instead.

    def __init__(self, toric_variety, polynomials):
        r"""
        See :class:`AlgebraicScheme_subscheme_toric` for documentation.

        TESTS::

            sage: P1xP1 = toric_varieties.P1xP1()
            sage: P1xP1.inject_variables()
            Defining s, t, x, y
            sage: import sage.schemes.toric.toric_subscheme as SCM
            sage: X = SCM.AlgebraicScheme_subscheme_toric(
            ....:       P1xP1, [x*s + y*t, x^3 + y^3])
            sage: X
            Closed subscheme of 2-d CPR-Fano toric variety
            covered by 4 affine patches defined by:
              s*x + t*y,
              x^3 + y^3
        """
        # Just to make sure that keyword arguments will be passed correctly
        super().__init__(toric_variety, polynomials)

    def _morphism(self, *args, **kwds):
        r"""
        Construct a morphism determined by action on points of ``self``.

        INPUT:

        - same as for
          :class:`~sage.schemes.toric.morphism.SchemeMorphism_polynomial_toric_variety`.

        OUTPUT: a :class:`~sage.schemes.toric.morphism.SchemeMorphism_polynomial_toric_variety`

        TESTS::

            sage: P1xP1 = toric_varieties.P1xP1()
            sage: P1xP1.inject_variables()
            Defining s, t, x, y
            sage: P1 = P1xP1.subscheme(s - t)
            sage: H = P1.Hom(P1xP1)
            sage: H([s, s, x, y])
            Scheme morphism:
              From: Closed subscheme of 2-d CPR-Fano toric variety
              covered by 4 affine patches defined by:
              s - t
              To:   2-d CPR-Fano toric variety covered by 4 affine patches
              Defn: Defined on coordinates by sending [s : t : x : y] to
                    [s : s : x : y]

            sage: sbar, tbar, xbar, ybar = P1.coordinate_ring().gens()                  # needs sage.libs.singular
            sage: P1._morphism(H, [sbar, sbar, xbar, ybar])                             # needs sage.libs.singular
            Scheme morphism:
              From: Closed subscheme of 2-d CPR-Fano toric variety
              covered by 4 affine patches defined by:
              s - t
              To:   2-d CPR-Fano toric variety covered by 4 affine patches
              Defn: Defined on coordinates by sending [s : t : x : y] to
                    [t : t : x : y]
        """
        from sage.schemes.toric.morphism import SchemeMorphism_polynomial_toric_variety
        return SchemeMorphism_polynomial_toric_variety(*args, **kwds)

    def _point_homset(self, *args, **kwds):
        r"""
        Construct a Hom-set for ``self``.

        INPUT:

        - same as for
          :class:`~sage.schemes.toric.homset.SchemeHomset_points_toric_field`.

        OUTPUT: a :class:`~sage.schemes.toric.homset.SchemeHomset_points_subscheme_toric_field`

        TESTS::

            sage: P2.<x,y,z> = toric_varieties.P2()
            sage: quadric = P2.subscheme([x^2 + y^2 + z^2])
            sage: quadric._point_homset(Spec(QQ), quadric)
            Set of rational points of Closed subscheme of 2-d CPR-Fano
            toric variety covered by 3 affine patches defined by:
              x^2 + y^2 + z^2
            sage: type(quadric.point_set())
            <class 'sage.schemes.toric.homset.SchemeHomset_points_subscheme_toric_field_with_category'>
        """
        from sage.schemes.toric.homset import SchemeHomset_points_subscheme_toric_field
        return SchemeHomset_points_subscheme_toric_field(*args, **kwds)

    def fan(self):
        """
        Return the fan of the ambient space.

        OUTPUT: a fan

        EXAMPLES::

            sage: P2.<x,y,z> = toric_varieties.P(2)
            sage: E = P2.subscheme([x^2 + y^2 + z^2])
            sage: E.fan()
            Rational polyhedral fan in 2-d lattice N
        """
        return self.ambient_space().fan()

    def affine_patch(self, i):
        r"""
        Return the ``i``-th affine patch of ``self`` as an affine
        toric algebraic scheme.

        INPUT:

        - ``i`` -- integer; index of a generating cone of the fan of the
          ambient space of ``self``

        OUTPUT:

        - subscheme of an affine :class:`toric variety
          <sage.schemes.toric.variety.ToricVariety_field>`
          corresponding to the pull-back of ``self`` by the embedding
          morphism of the ``i``-th :meth:`affine patch of the ambient
          space
          <sage.schemes.toric.variety.ToricVariety_field.affine_patch>`
          of ``self``.

        The result is cached, so the ``i``-th patch is always the same object
        in memory.

        EXAMPLES::

            sage: P1xP1 = toric_varieties.P1xP1()
            sage: patch1 = P1xP1.affine_patch(1)
            sage: patch1.embedding_morphism()
            Scheme morphism:
              From: 2-d affine toric variety
              To:   2-d CPR-Fano toric variety covered by 4 affine patches
              Defn: Defined on coordinates by sending [t : x] to [1 : t : x : 1]
            sage: P1xP1.inject_variables()
            Defining s, t, x, y
            sage: P1 = P1xP1.subscheme(x - y)
            sage: subpatch = P1.affine_patch(1)
            sage: subpatch
            Closed subscheme of 2-d affine toric variety defined by:
              x - 1
        """
        i = int(i)   # implicit type checking
        try:
            return self._affine_patches[i]
        except AttributeError:
            self._affine_patches = {}
        except KeyError:
            pass
        ambient_patch = self.ambient_space().affine_patch(i)
        phi_p = ambient_patch.embedding_morphism().defining_polynomials()
        patch = ambient_patch.subscheme(
                            [p(phi_p) for p in self.defining_polynomials()])
        patch._embedding_morphism = patch.hom(phi_p, self, check=False)
        self._affine_patches[i] = patch
        return patch

    def affine_algebraic_patch(self, cone=None, names=None):
        r"""
        Return the affine patch corresponding to ``cone`` as an affine
        algebraic scheme.

        INPUT:

        - ``cone`` -- a :class:`Cone
          <sage.geometry.cone.ConvexRationalPolyhedralCone>` `\sigma`
          of the fan. It can be omitted for an affine toric variety,
          in which case the single generating cone is used.

        OUTPUT:

        An :class:`affine algebraic subscheme
        <sage.schemes.affine.affine_subscheme.AlgebraicScheme_subscheme_affine>`
        corresponding to the patch `\mathop{Spec}(\sigma^\vee \cap M)`
        associated to the cone `\sigma`.

        See also :meth:`affine_patch`, which expresses the patches as
        subvarieties of affine toric varieties instead.

        REFERENCES:

        ..

            David A. Cox, "The Homogeneous Coordinate Ring of a Toric
            Variety", Lemma 2.2.
            :arxiv:`alg-geom/9210008v2`

        EXAMPLES::

            sage: P2.<x,y,z> = toric_varieties.P2()
            sage: cone = P2.fan().generating_cone(0)
            sage: V = P2.subscheme(x^3 + y^3 + z^3)
            sage: V.affine_algebraic_patch(cone)
            Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
              z0^3 + z1^3 + 1

            sage: # needs fpylll sage.libs.singular
            sage: cone = Cone([(0,1), (2,1)])
            sage: A2Z2.<x,y> = AffineToricVariety(cone)
            sage: A2Z2.affine_algebraic_patch()
            Closed subscheme of Affine Space of dimension 3 over Rational Field defined by:
              -z0*z1 + z2^2
            sage: V = A2Z2.subscheme(x^2 + y^2 - 1)
            sage: patch = V.affine_algebraic_patch();  patch
            Closed subscheme of Affine Space of dimension 3 over Rational Field defined by:
              -z0*z1 + z2^2,
              z0 + z1 - 1
            sage: nbhd_patch = V.neighborhood([1,0]).affine_algebraic_patch();  nbhd_patch
            Closed subscheme of Affine Space of dimension 3 over Rational Field defined by:
              -z0*z1 + z2^2,
              z0 + z1 - 1
            sage: nbhd_patch.embedding_center()
            (0, 1, 0)

        Here we got two defining equations. The first one describes
        the singularity of the ambient space and the second is the
        pull-back of `x^2+y^2-1` ::

            sage: lp = LatticePolytope([(1,0,0), (1,1,0), (1,1,1), (1,0,1), (-2,-1,-1)],
            ....:                      lattice=ToricLattice(3))
            sage: X.<x,y,u,v,t> = CPRFanoToricVariety(Delta_polar=lp)
            sage: Y = X.subscheme(x*v + y*u + t)
            sage: cone = Cone([(1,0,0), (1,1,0), (1,1,1), (1,0,1)])
            sage: Y.affine_algebraic_patch(cone)                                        # needs sage.libs.singular
            Closed subscheme of Affine Space of dimension 4 over Rational Field defined by:
              z0*z2 - z1*z3,
              z1 + z3 + 1
        """
        from sage.modules.free_module_element import vector
        from sage.misc.misc_c import prod
        ambient = self.ambient_space()
        fan = ambient.fan()
        if cone is None:
            assert ambient.is_affine()
            cone = fan.generating_cone(0)
        else:
            cone = fan.embed(cone)
        # R/I = C[sigma^dual cap M]
        R, I, dualcone = ambient._semigroup_ring(cone, names)

        # inhomogenize the Cox homogeneous polynomial with respect to the given cone
        inhomogenize = {ambient.coordinate_ring().gen(i): 1
                        for i in range(fan.nrays())
                        if i not in cone.ambient_ray_indices()}
        polynomials = [p.subs(inhomogenize) for p in self.defining_polynomials()]

        # map the monomial x^{D_m} to m, see reference.
        n_rho_matrix = cone.rays().matrix()

        def pullback_polynomial(p):
            result = R.zero()
            for coefficient, monomial in p:
                exponent = monomial.exponents()[0]
                exponent = [exponent[i] for i in cone.ambient_ray_indices()]
                exponent = vector(ZZ, exponent)
                m = n_rho_matrix.solve_right(exponent)
                assert all(x in ZZ for x in m), \
                    f'The polynomial {p} does not define a ZZ-divisor!'
                m_coeffs = dualcone.Hilbert_coefficients(m)
                result += coefficient * prod(R.gen(i)**m_coeffs[i]
                                             for i in range(R.ngens()))
            return result

        # construct the affine algebraic scheme to use as patch
        polynomials = [pullback_polynomial(_) for _ in polynomials]
        from sage.schemes.affine.affine_space import AffineSpace
        patch_cover = AffineSpace(R)
        polynomials = list(I.gens()) + polynomials
        polynomials = [x for x in polynomials if not x.is_zero()]
        patch = patch_cover.subscheme(polynomials)

        # TODO: If the cone is not smooth, then the coordinate_ring()
        # of the affine toric variety is wrong; it should be the
        # G-invariant part. So we can't construct the embedding
        # morphism in that case.
        if cone.is_smooth():
            x = ambient.coordinate_ring().gens()
            phi = []
            for i in range(fan.nrays()):
                if i in cone.ambient_ray_indices():
                    phi.append(pullback_polynomial(x[i]))
                else:
                    phi.append(1)
            patch._embedding_morphism = patch.hom(phi, self)
        else:
            patch._embedding_morphism = (NotImplementedError,
               'I only know how to construct embedding morphisms for smooth patches')

        try:
            point = self.embedding_center()
        except AttributeError:
            return patch

        # it remains to find the preimage of point
        # map m to the monomial x^{D_m}, see reference.
        F = ambient.coordinate_ring().fraction_field()
        image = [prod([F.gen(i)**(m * n)
                       for i, n in enumerate(fan.rays())])
                 for m in dualcone.Hilbert_basis()]
        patch._embedding_center = tuple(f(list(point)) for f in image)
        return patch

    def _best_affine_patch(self, point):
        r"""
        Return the best affine patch of the ambient toric variety.

        INPUT:

        - ``point`` -- a point of the algebraic subscheme

        OUTPUT: integer. The index of the patch. See :meth:`affine_patch`

        EXAMPLES::

            sage: P.<x,y,z> = toric_varieties.P2()
            sage: S = P.subscheme(x + 2*y + 3*z)
            sage: S._best_affine_patch(P.point([2,-3,0]))
            1
            sage: S._best_affine_patch([2,-3,0])
            1
        """
        # TODO: this method should pick a "best" patch in the sense
        # that it is numerically stable to dehomogenize, see the
        # corresponding method for projective varieties.
        point = list(point)
        zeros = {i for i, coord in enumerate(point) if coord == 0}
        for cone_idx, cone in enumerate(self.ambient_space().fan().generating_cones()):
            if zeros.issubset(cone.ambient_ray_indices()):
                return cone_idx
        assert False, 'The point must not have been a point of the toric variety.'

    def neighborhood(self, point):
        r"""
        Return a toric algebraic scheme isomorphic to neighborhood of
        the ``point``.

        INPUT:

        - ``point`` -- a point of the toric algebraic scheme

        OUTPUT:

        An affine toric algebraic scheme (polynomial equations in an
        affine toric variety) with fixed
        :meth:`~AlgebraicScheme.embedding_morphism` and
        :meth:`~AlgebraicScheme.embedding_center`.

        EXAMPLES::

            sage: # needs sage.libs.singular
            sage: P.<x,y,z> = toric_varieties.P2()
            sage: S = P.subscheme(x + 2*y + 3*z)
            sage: s = S.point([0,-3,2]); s
            [0 : -3 : 2]
            sage: patch = S.neighborhood(s); patch
            Closed subscheme of 2-d affine toric variety defined by:
              x + 2*y + 6
            sage: patch.embedding_morphism()
            Scheme morphism:
              From: Closed subscheme of 2-d affine toric variety defined by: x + 2*y + 6
              To:   Closed subscheme of 2-d CPR-Fano toric variety
                    covered by 3 affine patches defined by: x + 2*y + 3*z
              Defn: Defined on coordinates by sending [x : y] to [-2*y - 6 : y : 2]
            sage: patch.embedding_center()
            [0 : -3]
            sage: patch.embedding_morphism()(patch.embedding_center())
            [0 : -3 : 2]

        A more complicated example::

            sage: # needs sage.libs.singular
            sage: dP6.<x0,x1,x2,x3,x4,x5> = toric_varieties.dP6()
            sage: twoP1 = dP6.subscheme(x0*x3)
            sage: patch = twoP1.neighborhood([0,1,2, 3,4,5]); patch
            Closed subscheme of 2-d affine toric variety defined by:
              3*x0
            sage: patch.embedding_morphism()
            Scheme morphism:
              From: Closed subscheme of 2-d affine toric variety defined by: 3*x0
              To:   Closed subscheme of 2-d CPR-Fano toric variety
                    covered by 6 affine patches defined by: x0*x3
              Defn: Defined on coordinates by sending [x0 : x1] to [0 : x1 : 2 : 3 : 4 : 5]
            sage: patch.embedding_center()
            [0 : 1]
            sage: patch.embedding_morphism()(patch.embedding_center())
            [0 : 1 : 2 : 3 : 4 : 5]
        """
        point = list(point)
        self._check_satisfies_equations(point)
        PP = self.ambient_space()
        fan = PP.fan()
        cone_idx = self._best_affine_patch(point)
        cone = fan.generating_cone(cone_idx)

        patch_cover = PP.affine_patch(cone_idx)
        R = patch_cover.coordinate_ring()
        phi = []
        point_preimage = []
        for i in range(fan.nrays()):
            try:
                ray_index = cone.ambient_ray_indices().index(i)
                phi.append(R.gen(ray_index))
                point_preimage.append(point[i])
            except ValueError:
                phi.append(point[i])
        pullback_polys = [f(phi) for f in self.defining_polynomials()]
        patch = patch_cover.subscheme(pullback_polys)
        S = patch.coordinate_ring()
        phi_reduced = [S(t) for t in phi]

        patch._embedding_center = patch(point_preimage)
        patch._embedding_morphism = patch.hom(phi_reduced, self)
        return patch

    def dimension(self):
        """
        Return the dimension of ``self``.

        OUTPUT: integer; if ``self`` is empty, `-1` is returned

        EXAMPLES::

            sage: # needs sage.libs.singular
            sage: P1xP1 = toric_varieties.P1xP1()
            sage: P1xP1.inject_variables()
            Defining s, t, x, y
            sage: P1 = P1xP1.subscheme(s - t)
            sage: P1.dimension()
            1
            sage: P1xP1.subscheme([s - t, (s-t)^2]).dimension()
            1
            sage: P1xP1.subscheme([s, t]).dimension()
            -1
        """
        if '_dimension' in self.__dict__:
            return self._dimension
        npatches = self.ambient_space().fan().ngenerating_cones()
        dims = [self.affine_patch(i).dimension() for i in range(npatches)]
        self._dimension = max(dims)
        return self._dimension

    def is_smooth(self, point=None):
        r"""
        Test whether the algebraic subscheme is smooth.

        INPUT:

        - ``point`` -- a point or ``None`` (default); the point to
          test smoothness at

        OUTPUT:

        boolean; if no point was specified, returns whether the
        algebraic subscheme is smooth everywhere. Otherwise,
        smoothness at the specified point is tested.

        EXAMPLES::

            sage: # needs sage.libs.singular
            sage: P2.<x,y,z> = toric_varieties.P2()
            sage: cuspidal_curve = P2.subscheme([y^2*z - x^3])
            sage: cuspidal_curve
            Closed subscheme of 2-d CPR-Fano toric variety covered by 3 affine patches defined by:
              -x^3 + y^2*z
            sage: cuspidal_curve.is_smooth([1,1,1])
            True
            sage: cuspidal_curve.is_smooth([0,0,1])
            False
            sage: cuspidal_curve.is_smooth()
            False

        Any sufficiently generic cubic hypersurface is smooth::

            sage: P2.subscheme([y^2*z-x^3+z^3+1/10*x*y*z]).is_smooth()                  # needs sage.libs.singular
            True

        A more complicated example::

            sage: # needs sage.libs.singular
            sage: dP6.<x0,x1,x2,x3,x4,x5> = toric_varieties.dP6()
            sage: disjointP1s = dP6.subscheme(x0*x3)
            sage: disjointP1s.is_smooth()
            True
            sage: intersectingP1s = dP6.subscheme(x0*x1)
            sage: intersectingP1s.is_smooth()
            False

        A smooth hypersurface in a compact singular toric variety::

            sage: # needs sage.libs.singular
            sage: lp = LatticePolytope([(1,0,0), (1,1,0), (1,1,1), (1,0,1), (-2,-1,-1)],
            ....:                      lattice=ToricLattice(3))
            sage: X.<x,y,u,v,t> = CPRFanoToricVariety(Delta_polar=lp)
            sage: Y = X.subscheme(x*v + y*u + t)
            sage: cone = Cone([(1,0,0), (1,1,0), (1,1,1), (1,0,1)])
            sage: Y.is_smooth()
            True
        """
        if point is not None:
            toric_patch = self.neighborhood(point)
            return toric_patch.is_smooth(toric_patch.embedding_center())

        # testing smoothness everywhere tends to be expensive
        if '_smooth' in self.__dict__:
            return self._smooth
        npatches = self.ambient_space().fan().ngenerating_cones()
        self._smooth = all(self.affine_patch(i).is_smooth()
                           for i in range(npatches))
        return self._smooth

    def is_nondegenerate(self):
        r"""
        Check if ``self`` is nondegenerate.

        OUTPUT:

        Whether the variety is nondegenerate, that is, the intersection
        with every open torus orbit is smooth and transversal.

        EXAMPLES::

            sage: P2.<x,y,z> = toric_varieties.P2()
            sage: P2.subscheme([x^3 + y^3 + z^3]).is_nondegenerate()
            True
            sage: P2.subscheme([x*y*z]).is_nondegenerate()
            False
            sage: X = P2.subscheme([(x-y)^2*(x+y) + x*y*z + z^3])
            sage: X.is_smooth()
            True
            sage: X.is_nondegenerate()
            False

        A K3 surface in `\mathbf{P}^1 \times \mathbf{P}^1 \times \mathbf{P}^1`::

            sage: diamond = lattice_polytope.cross_polytope(3)
            sage: fan = FaceFan(diamond)
            sage: P1xP1xP1 = ToricVariety(fan)
            sage: z0, z1, z2, z3, z4, z5 = P1xP1xP1.gens()
            sage: t = 5
            sage: F = z0^2*z1^2*z2^2 + z1^2*z2^2*z3^2 + z0^2*z2^2*z4^2\
            ....: + z2^2*z3^2*z4^2 + t*z0*z1*z2*z3*z4*z5 + z0^2*z1^2*z5^2\
            ....: + z1^2*z3^2*z5^2 + z0^2*z4^2*z5^2 + z3^2*z4^2*z5^2
            sage: X = P1xP1xP1.subscheme([F])
            sage: X.is_smooth()
            True
            sage: X.is_nondegenerate()
            False

        Taking a random change of variables breaks the symmetry, but
        makes the surface nondegenerate::

            sage: F1 = F.subs(z0=1*z0 + 1*z3, z3=1*z0 + 2*z3,
            ....:             z1=-2*z1 + -1*z4, z4=1*z1 + 2*z4,
            ....:             z2=-3*z2 + -1*z5, z5=-3*z2 + 2*z5)
            sage: Y = P1xP1xP1.subscheme([F1])
            sage: Y.is_smooth()
            True
            sage: Y.is_nondegenerate()
            True

        This example is from Hamm, :arxiv:`1106.1826v1`. It addresses
        an issue raised at :issue:`15239`::

            sage: X = toric_varieties.WP([1,4,2,3], names='z0 z1 z2 z3')
            sage: X.inject_variables()
            Defining z0, z1, z2, z3
            sage: g0 = z1^3 + z2^6 + z3^4
            sage: g = g0 - 2*z3^2*z0^6 + z2*z0^10 + z0^12
            sage: Y = X.subscheme([g])
            sage: Y.is_nondegenerate()
            False

        It handles nonzero characteristic::

            sage: P2.<x,y,z> = toric_varieties.P2()
            sage: f = x^5 + 2*x*y^4 + y^5 - 2*y^3*z^2 + x*z^4 - 2*z^5
            sage: P2.change_ring(GF(5)).subscheme([f]).is_nondegenerate()
            True
            sage: P2.change_ring(GF(7)).subscheme([f]).is_nondegenerate()
            False

        TESTS:

        Some corner cases discussed at :issue:`15239`::

            sage: P2.<x,y,z> = toric_varieties.P2()
            sage: P2.subscheme([]).is_nondegenerate()
            False
            sage: P2.subscheme([x]).is_nondegenerate()
            False
        """
        X = self.ambient_space()
        fan = X.fan()
        SR = X.Stanley_Reisner_ideal()
        from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
        R = PolynomialRing(X.base_ring(), fan.nrays() + SR.ngens(), 't')
        slack = R.gens()[fan.nrays():]
        SR = SR.change_ring(R)

        def restrict(cone):
            patch = {}
            divide = {}
            for i in cone.ambient_ray_indices():
                patch[R.gen(i)] = R.zero()   # restrict to torus orbit
                # divide out highest power of R.gen(i)
                divide[R.gen(i)] = R.one()
            ideal = self.defining_ideal().change_ring(R)
            ideal = ideal.subs(patch)
            mat = jacobian(ideal.gens(), R.gens()[:fan.nrays()])
            minors = mat.minors(self.codimension())
            minors = tuple([ideal.reduce(m) for m in minors])
            Jac_patch = R.ideal(ideal.gens() + minors)
            SR_patch = R.ideal([monomial * slack[i] - R.one()
                                for i, monomial in
                                enumerate(SR.subs(divide).gens())])
            return ideal, Jac_patch + SR_patch

        for dim in range(fan.dim() + 1):
            for cone in fan(dim):
                ideal1, ideal2 = restrict(cone)
                if ideal1.is_zero() or ideal2.dimension() != -1:
                    return False

        return True

    def is_schon(self):
        r"""
        Check if ``self`` is schon (nondegenerate).

        See :meth:`is_nondegenerate` for further documentation.

        EXAMPLES::

            sage: # needs sage.libs.singular
            sage: P2.<x,y,z> = toric_varieties.P2()
            sage: X = P2.subscheme([(x-y)^2*(x+y) + x*y*z + z^3])
            sage: X.is_smooth()
            True
            sage: X.is_schon()
            False
        """
        return self.is_nondegenerate()


class AlgebraicScheme_subscheme_affine_toric(AlgebraicScheme_subscheme_toric):
    r"""
    Construct an algebraic subscheme of an affine toric variety.

    .. WARNING::

        You should not create objects of this class directly. The preferred
        method to construct such subschemes is to use
        :meth:`~ToricVariety_field.subscheme` method of
        :class:`toric varieties <ToricVariety_field>`.

    INPUT:

    - ``toric_variety`` -- ambient :class:`affine toric variety
      <ToricVariety_field>`

    - ``polynomials`` -- single polynomial, list, or ideal of defining
      polynomials in the coordinate ring of ``toric_variety``

    OUTPUT:

    A :class:`algebraic subscheme of an affine toric variety
    <AlgebraicScheme_subscheme_affine_toric>`.

    TESTS::

        sage: P1xP1 = toric_varieties.P1xP1()
        sage: P1xP1.inject_variables()
        Defining s, t, x, y
        sage: import sage.schemes.toric.toric_subscheme as SCM
        sage: X = SCM.AlgebraicScheme_subscheme_toric(
        ....:       P1xP1, [x*s + y*t, x^3 + y^3])
        sage: X
        Closed subscheme of 2-d CPR-Fano toric variety
        covered by 4 affine patches defined by:
          s*x + t*y,
          x^3 + y^3

    A better way to construct the same scheme as above::

        sage: P1xP1.subscheme([x*s + y*t, x^3 + y^3])
        Closed subscheme of 2-d CPR-Fano toric variety
        covered by 4 affine patches defined by:
          s*x + t*y,
          x^3 + y^3
    """

    def __init__(self, toric_variety, polynomials):
        r"""
        See :class:`AlgebraicScheme_subscheme_toric` for documentation.

        TESTS::

            sage: P1xP1 = toric_varieties.P1xP1()
            sage: P1xP1.inject_variables()
            Defining s, t, x, y
            sage: import sage.schemes.toric.toric_subscheme as SCM
            sage: X = SCM.AlgebraicScheme_subscheme_toric(
            ....:       P1xP1, [x*s + y*t, x^3 + y^3])
            sage: X
            Closed subscheme of 2-d CPR-Fano toric variety
            covered by 4 affine patches defined by:
              s*x + t*y,
              x^3 + y^3
        """
        assert toric_variety.is_affine(), 'The toric variety must be affine!'
        # Just to make sure that keyword arguments will be passed correctly
        super().__init__(toric_variety, polynomials)

    def dimension(self):
        """
        Return the dimension of ``self``.

        OUTPUT: integer

        EXAMPLES::

            sage: # needs sage.libs.singular
            sage: P1xP1.<s0,s1,t0,t1> = toric_varieties.P1xP1()
            sage: P1 = P1xP1.subscheme(s0 - s1)
            sage: P1.dimension()
            1

        A more complicated example where the ambient toric variety is
        not smooth::

            sage: # needs sage.libs.singular
            sage: X.<x,y> = toric_varieties.A2_Z2()
            sage: X.is_smooth()
            False
            sage: Y = X.subscheme([x*y, x^2]); Y
            Closed subscheme of 2-d affine toric variety defined by:
              x*y,
              x^2
            sage: Y.dimension()
            1
        """
        if '_dimension' in self.__dict__:
            return self._dimension

        if self.ambient_space().is_smooth():
            self._dimension = self.defining_ideal().dimension()
        else:
            self._dimension = self.affine_algebraic_patch().dimension()
        return self._dimension

    def is_smooth(self, point=None):
        r"""
        Test whether the algebraic subscheme is smooth.

        INPUT:

        - ``point`` -- a point or ``None`` (default); the point to
          test smoothness at

        OUTPUT:

        boolean; if no point was specified, returns whether the
        algebraic subscheme is smooth everywhere. Otherwise,
        smoothness at the specified point is tested.

        EXAMPLES::

            sage: # needs sage.libs.singular
            sage: A2.<x,y> = toric_varieties.A2()
            sage: cuspidal_curve = A2.subscheme([y^2 - x^3])
            sage: cuspidal_curve
            Closed subscheme of 2-d affine toric variety defined by:
              -x^3 + y^2
            sage: cuspidal_curve.is_smooth([1,1])
            True
            sage: cuspidal_curve.is_smooth([0,0])
            False
            sage: cuspidal_curve.is_smooth()
            False
            sage: circle = A2.subscheme(x^2 + y^2 - 1)
            sage: circle.is_smooth([1,0])
            True
            sage: circle.is_smooth()
            True

        A more complicated example where the ambient toric variety is
        not smooth::

            sage: # needs sage.libs.singular
            sage: X.<x,y> = toric_varieties.A2_Z2()    # 2-d affine space mod Z/2
            sage: X.is_smooth()
            False
            sage: Y = X.subscheme([x*y, x^2])   # (twice the x=0 curve) mod Z/2
            sage: Y
            Closed subscheme of 2-d affine toric variety defined by:
              x*y,
              x^2
            sage: Y.dimension()   # Y is a Weil divisor but not Cartier
            1
            sage: Y.is_smooth()
            True
            sage: Y.is_smooth([0,0])
            True
        """
        if point is not None:
            self._check_satisfies_equations(point)
            if self.ambient_space().is_smooth():
                R = self.ambient_space().coordinate_ring()
                point_subs = dict(zip(R.gens(), point))
                Jac = self.Jacobian().subs(point_subs)
                return not Jac.is_zero()
            else:
                self._embedding_center = self.point(point)
                affine = self.affine_algebraic_patch()
                return affine.is_smooth(affine.embedding_center())

        # testing smoothness everywhere tends to be expensive
        if '_smooth' in self.__dict__:
            return self._smooth

        if self.ambient_space().is_smooth():
            sing_dim = self.Jacobian().dimension()
            self._smooth = (sing_dim == -1)
        else:
            self._smooth = self.affine_algebraic_patch().is_smooth()

        return self._smooth



# sage_setup: distribution = sagemath-polyhedra
cimport cython
from sage.data_structures.list_of_pairs cimport ListOfPairs
from sage.structure.sage_object         cimport SageObject
from sage.geometry.polyhedron.combinatorial_polyhedron.face_iterator                     cimport FaceIterator, CombinatorialFace
from sage.geometry.polyhedron.combinatorial_polyhedron.list_of_faces                     cimport ListOfFaces
from sage.geometry.polyhedron.combinatorial_polyhedron.face_data_structure               cimport face_t
from sage.geometry.polyhedron.combinatorial_polyhedron.polyhedron_face_lattice           cimport PolyhedronFaceLattice


@cython.final
cdef class CombinatorialPolyhedron(SageObject):
    cdef public dict _cached_methods

    # Do not assume any of those attributes to be initialized, use the corresponding methods instead.
    cdef tuple _Vrep                       # the names of VRep, if they exist
    cdef tuple _facet_names                # the names of HRep without equations, if they exist
    cdef tuple _equations                  # stores equations, given on input (might belong to Hrep)
    cdef int _dimension                    # stores dimension, -2 on init
    cdef unsigned int _n_Hrepresentation   # Hrep might include equations
    cdef unsigned int _n_Vrepresentation   # Vrep might include rays/lines
    cdef size_t _n_facets                  # length Hrep without equations
    cdef bint _bounded                     # ``True`` iff Polyhedron is bounded
    cdef ListOfFaces _bitrep_facets        # facets in bit representation
    cdef ListOfFaces _bitrep_Vrep          # vertices in bit representation
    cdef face_t _far_face                  # a 'face' containing all none-vertices of Vrep
    cdef tuple _far_face_tuple
    cdef tuple _f_vector

    cdef ListOfPairs _edges                    # stores edges labeled by vertex indices
    cdef ListOfPairs _ridges                   # stores ridges labeled by facet indices
    cdef ListOfPairs _face_lattice_incidences  # stores incidences in Hasse diagram labeled indices of the faces
    cdef PolyhedronFaceLattice _all_faces     # class to generate Hasse diagram incidences

    cdef tuple Vrep(self)
    cdef tuple facet_names(self)
    cdef tuple equations(self)
    cdef unsigned int n_Vrepresentation(self) noexcept
    cdef unsigned int n_Hrepresentation(self) noexcept
    cdef bint is_bounded(self) noexcept
    cdef ListOfFaces bitrep_facets(self)
    cdef ListOfFaces bitrep_Vrep(self)
    cdef tuple far_face_tuple(self)
    cdef int _algorithm_to_dual(self, algorithm) except -2

    # Methods to initialize the combinatorial polyhedron.
    cdef _init_from_polyhedron(self, data)
    cdef _init_from_lattice_polytope(self, data)
    cdef _init_from_cone(self, data)
    cdef _init_facet_names(self, facets)
    cdef _init_from_incidence_matrix(self, data)
    cdef _init_from_list_of_facets(self, data)
    cdef _init_from_ListOfFaces(self, ListOfFaces facets, ListOfFaces Vrep)
    cdef _initialize_far_face(self)
    cdef _init_as_trivial_polyhedron(self, int dimension)

    # Methods to obtain a different combinatorial polyhedron.
    cpdef CombinatorialPolyhedron dual(self)
    cpdef CombinatorialPolyhedron pyramid(self, new_vertex=*, new_facet=*)

    cdef FaceIterator _face_iter(self, bint dual, int dimension)
    cdef int _compute_f_vector(self, size_t num_threads, size_t parallelization_depth, int dual) except -1
    cdef int _persist_f_vector(self, size_t* input_f_vector, bint input_is_reversed) except -1

    cdef inline int _compute_edges(self, dual) except -1:
        return self._compute_edges_or_ridges(dual, True)

    cdef inline int _compute_ridges(self, dual) except -1:
        return self._compute_edges_or_ridges(dual, False)

    cdef int _compute_edges_or_ridges(self, int dual, bint do_edges) except -1
    cdef size_t _compute_edges_or_ridges_with_iterator(
            self, FaceIterator face_iter, const bint do_atom_rep,
            ListOfPairs edges, size_t* f_vector) except -1

    cdef int _compute_face_lattice_incidences(self) except -1
